"""Genetic Programming in Python, with a scikit-learn inspired API

``gplearn`` is a set of algorithms for learning genetic programming models.

"""
__version__ = '0.4.2'

__all__ = ['genetic', 'functions', 'fitness']
